from __future__ import annotations

__all__ = ["ValuesIndicesTuple"]

from collections import namedtuple
from typing import Any

from coola import objects_are_equal
from coola.utils import str_indent, str_mapping


class ValuesIndicesTuple(namedtuple("TupleValuesIndices", ["values", "indices"])):
    r"""Implements a namedtuple to represent the pair ``(values,
    indices)``.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from redcat.return_types import ValuesIndicesTuple
        >>> x = ValuesIndicesTuple(torch.ones(6), torch.arange(6))
        >>> x
        ValuesIndicesTuple(
          (values): tensor([1., 1., 1., 1., 1., 1.])
          (indices): tensor([0, 1, 2, 3, 4, 5])
        )
        >>> x.values
        tensor([1., 1., 1., 1., 1., 1.])
        >>> x.indices
        tensor([0, 1, 2, 3, 4, 5])
    """

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ValuesIndicesTuple):
            return False
        return objects_are_equal(self.values, other.values) and objects_are_equal(
            self.indices, other.indices
        )

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"values": self.values, "indices": self.indices}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"



from __future__ import annotations

__all__ = [
    "BatchedTensorSeq",
    "check_data_and_dims",
    "from_sequences",
]

from collections.abc import Callable, Iterable, Sequence
from typing import Any

import numpy as np
import torch
from coola import objects_are_allclose, objects_are_equal
from torch import Tensor
from torch.nn.utils.rnn import pad_sequence

from redcat import BaseBatch, tensor
from redcat.tensor import BatchedTensor, get_batch_dims
from redcat.utils.common import check_batch_dims, check_seq_dims, get_seq_dims
from redcat.utils.tensor import (
    align_to_batch_seq,
    align_to_seq_batch,
    compute_batch_seq_permutation,
)

HANDLED_FUNCTIONS = {
    torch.amax: tensor.amax,
    torch.amin: tensor.amin,
    torch.argmax: tensor.argmax,
    torch.argmin: tensor.argmin,
    torch.argsort: tensor.argsort,
    torch.cat: tensor.cat,
    torch.chunk: tensor.chunk,
    torch.max: tensor.torchmax,
    torch.maximum: tensor.maximum,
    torch.mean: tensor.mean,
    torch.median: tensor.median,
    torch.min: tensor.torchmin,
    torch.minimum: tensor.minimum,
    torch.nanmean: tensor.nanmean,
    torch.nanmedian: tensor.nanmedian,
    torch.nansum: tensor.nansum,
    torch.prod: tensor.prod,
    torch.select: tensor.select,
    torch.sort: tensor.sort,
    torch.split: tensor.split,
    torch.sum: tensor.torchsum,
    torch.take_along_dim: tensor.take_along_dim,
}


class BatchedTensorSeq(BatchedTensor):
    r"""Implements a batched tensor to easily manipulate a batch of
    sequences.

    Args:
    ----
        data (array_like): Specifies the data for the tensor. It can
            be a torch.Tensor, list, tuple, NumPy ndarray, scalar,
            and other types.
        batch_dim (int, optional): Specifies the batch dimension
            in the ``torch.Tensor`` object. Default: ``0``
        seq_dim (int, optional): Specifies the sequence dimension in
            the ``torch.Tensor`` object. Default: ``1``
        kwargs: Keyword arguments that are passed to
            ``torch.as_tensor``.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from redcat import BatchedTensorSeq
        >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5))
    """

    def __init__(self, data: Any, *, batch_dim: int = 0, seq_dim: int = 1, **kwargs) -> None:
        super().__init__(data, batch_dim=batch_dim, **kwargs)
        check_data_and_dims(self._data, batch_dim, seq_dim)
        self._seq_dim = int(seq_dim)

    def __repr__(self) -> str:
        return repr(self._data)[:-1] + f", batch_dim={self._batch_dim}, seq_dim={self._seq_dim})"

    @classmethod
    def __torch_function__(
        cls,
        func: Callable,
        types: tuple[type, ...],
        args: tuple[Any, ...] = (),
        kwargs: dict[str, Any] | None = None,
    ) -> BatchedTensorSeq:
        kwargs = kwargs or {}
        if handled_func := HANDLED_FUNCTIONS.get(func, None):
            return handled_func(*args, **kwargs)

        batch_dims = get_batch_dims(args, kwargs)
        check_batch_dims(batch_dims)
        seq_dims = get_seq_dims(args, kwargs)
        check_seq_dims(seq_dims)
        args = [a._data if hasattr(a, "_data") else a for a in args]
        return cls(func(*args, **kwargs), batch_dim=batch_dims.pop(), seq_dim=seq_dims.pop())

    @property
    def seq_dim(self) -> int:
        r"""int: The sequence dimension in the ``torch.Tensor`` object."""
        return self._seq_dim

    @property
    def seq_len(self) -> int:
        r"""``int``: The sequence length."""
        return self._data.shape[self._seq_dim]

    ###############################
    #     Creation operations     #
    ###############################

    def new_full(
        self,
        fill_value: float | int | bool,
        batch_size: int | None = None,
        seq_len: int | None = None,
        **kwargs,
    ) -> BatchedTensorSeq:
        r"""Creates a batch filled with a scalar value.

        By default, the tensor in the returned batch has the same
        shape, ``torch.dtype`` and ``torch.device`` as the tensor in
        the current batch.

        Args:
        ----
            fill_value (float or int or bool): Specifies the number
                to fill the batch with.
            batch_size (int or ``None``): Specifies the batch size.
                If ``None``, the batch size of the current batch is
                used. Default: ``None``.
            seq_len (int or ``None``): Specifies the sequence length.
                If ``None``, the sequence length of the current batch
                is used. Default: ``None``.
            **kwargs: See the documentation of
                ``torch.Tensor.new_full``.

        Returns:
        -------
            ``BaseBatchedTensor``: A batch filled with the scalar value.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.ones(2, 3))
            >>> batch.new_full(42)
            tensor([[42., 42., 42.],
                    [42., 42., 42.]], batch_dim=0, seq_dim=1)
            >>> batch.new_full(42, batch_size=5)
            tensor([[42., 42., 42.],
                    [42., 42., 42.],
                    [42., 42., 42.],
                    [42., 42., 42.],
                    [42., 42., 42.]], batch_dim=0, seq_dim=1)
            >>> batch.new_full(42, seq_len=5)
            tensor([[42., 42., 42., 42., 42.],
                    [42., 42., 42., 42., 42.]], batch_dim=0, seq_dim=1)
        """
        shape = list(self._data.shape)
        if batch_size is not None:
            shape[self._batch_dim] = batch_size
        if seq_len is not None:
            shape[self._seq_dim] = seq_len
        kwargs["dtype"] = kwargs.get("dtype", self.dtype)
        kwargs["device"] = kwargs.get("device", self.device)
        return self._create_new_batch(torch.full(size=shape, fill_value=fill_value, **kwargs))

    def new_ones(
        self,
        batch_size: int | None = None,
        seq_len: int | None = None,
        **kwargs,
    ) -> BatchedTensorSeq:
        r"""Creates a batch filled with the scalar value ``1``.

        By default, the tensor in the returned batch has the same
        shape, ``torch.dtype`` and ``torch.device`` as the tensor in
        the current batch.

        Args:
        ----
            batch_size (int or ``None``): Specifies the batch size.
                If ``None``, the batch size of the current batch is
                used. Default: ``None``.
            seq_len (int or ``None``): Specifies the sequence length.
                If ``None``, the sequence length of the current batch
                is used. Default: ``None``.
            **kwargs: See the documentation of
                ``torch.Tensor.new_ones``.

        Returns:
        -------
            ``BaseBatchedTensor``: A batch filled with the scalar
                value ``1``.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.zeros(2, 3))
            >>> batch.new_ones()
            tensor([[1., 1., 1.],
                    [1., 1., 1.]], batch_dim=0, seq_dim=1)
            >>> batch.new_ones(batch_size=5)
            tensor([[1., 1., 1.],
                    [1., 1., 1.],
                    [1., 1., 1.],
                    [1., 1., 1.],
                    [1., 1., 1.]], batch_dim=0, seq_dim=1)
            >>> batch.new_ones(seq_len=5)
            tensor([[1., 1., 1., 1., 1.],
                    [1., 1., 1., 1., 1.]], batch_dim=0, seq_dim=1)
        """
        shape = list(self._data.shape)
        if batch_size is not None:
            shape[self._batch_dim] = batch_size
        if seq_len is not None:
            shape[self._seq_dim] = seq_len
        kwargs["dtype"] = kwargs.get("dtype", self.dtype)
        kwargs["device"] = kwargs.get("device", self.device)
        return self._create_new_batch(torch.ones(*shape, **kwargs))

    def new_zeros(
        self,
        batch_size: int | None = None,
        seq_len: int | None = None,
        **kwargs,
    ) -> BatchedTensorSeq:
        r"""Creates a batch filled with the scalar value ``0``.

        By default, the tensor in the returned batch has the same
        shape, ``torch.dtype`` and ``torch.device`` as the tensor
        in the current batch.

        Args:
        ----
            batch_size (int or ``None``): Specifies the batch size.
                If ``None``, the batch size of the current batch is
                used. Default: ``None``.
            seq_len (int or ``None``): Specifies the sequence length.
                If ``None``, the sequence length of the current batch
                is used. Default: ``None``.
            **kwargs: See the documentation of
                ``torch.Tensor.new_zeros``.

        Returns:
        -------
            ``BaseBatchedTensor``: A batch filled with the scalar
                value ``0``.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.ones(2, 3))
            >>> batch.new_zeros()
            tensor([[0., 0., 0.],
                    [0., 0., 0.]], batch_dim=0, seq_dim=1)
            >>> batch.new_zeros(batch_size=5)
            tensor([[0., 0., 0.],
                    [0., 0., 0.],
                    [0., 0., 0.],
                    [0., 0., 0.],
                    [0., 0., 0.]], batch_dim=0, seq_dim=1)
            >>> batch.new_zeros(seq_len=5)
            tensor([[0., 0., 0., 0., 0.],
                    [0., 0., 0., 0., 0.]], batch_dim=0, seq_dim=1)
        """
        shape = list(self._data.shape)
        if batch_size is not None:
            shape[self._batch_dim] = batch_size
        if seq_len is not None:
            shape[self._seq_dim] = seq_len
        kwargs["dtype"] = kwargs.get("dtype", self.dtype)
        kwargs["device"] = kwargs.get("device", self.device)
        return self._create_new_batch(torch.zeros(*shape, **kwargs))

    @classmethod
    def from_seq_batch(cls, data: Any, **kwargs) -> BatchedTensorSeq:
        r"""Creates a batch where the first dimension is the sequence
        dimension and the second dimension is the batch dimension.

        Args:
        ----
            data (array_like): Specifies the data for the tensor. It can
                be a torch.Tensor, list, tuple, NumPy ndarray, scalar,
                and other types.
            kwargs: Keyword arguments that are passed to
                ``torch.as_tensor``.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq.from_seq_batch(torch.ones(2, 3))
            >>> batch
            tensor([[1., 1., 1.],
                    [1., 1., 1.]], batch_dim=1, seq_dim=0)
        """
        return cls(data, batch_dim=1, seq_dim=0, **kwargs)

    #################################
    #     Comparison operations     #
    #################################

    def allclose(
        self, other: Any, rtol: float = 1e-5, atol: float = 1e-8, equal_nan: bool = False
    ) -> bool:
        if not isinstance(other, BatchedTensorSeq):
            return False
        if self._batch_dim != other.batch_dim or self._seq_dim != other.seq_dim:
            return False
        if self._data.shape != other.data.shape:
            return False
        return objects_are_allclose(
            self._data, other.data, rtol=rtol, atol=atol, equal_nan=equal_nan
        )

    def equal(self, other: Any) -> bool:
        if not isinstance(other, BatchedTensorSeq):
            return False
        if self._batch_dim != other.batch_dim or self._seq_dim != other.seq_dim:
            return False
        return objects_are_equal(self._data, other.data)

    ##################################################
    #     Mathematical | arithmetical operations     #
    ##################################################

    ###########################################################
    #     Mathematical | advanced arithmetical operations     #
    ###########################################################

    def argsort_along_seq(self, *args, **kwargs) -> BatchedTensorSeq:
        r"""Sorts the elements of the batch along the sequence dimension
        in monotonic order by value.

        Args:
        ----
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
        -------
            ``BatchedTensor``: The indices that sort the batch along
                the sequence dimension.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5))
            >>> batch.argsort_along_seq(descending=True)
            tensor([[4, 3, 2, 1, 0],
                    [4, 3, 2, 1, 0]], batch_dim=0, seq_dim=1)
        """
        return self.argsort(self._seq_dim, *args, **kwargs)

    def cumsum_along_seq(self, **kwargs) -> BatchedTensorSeq:
        r"""Computes the cumulative sum of elements of the current batch
        in the sequence dimension.

        Args:
        ----
            **kwargs: see ``torch.cumsum`` documentation

        Returns:
        -------
            ``BatchedTensorSeq``: A batch with the cumulative sum of
                elements of the current batch in the sequence dimension.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5)).cumsum_along_seq()
            >>> batch
            tensor([[ 0,  1,  3,  6, 10],
                    [ 5, 11, 18, 26, 35]], batch_dim=0, seq_dim=1)
        """
        return self.cumsum(dim=self._seq_dim, **kwargs)

    def cumsum_along_seq_(self) -> None:
        r"""Computes the cumulative sum of elements of the current batch
        in the sequence dimension.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5))
            >>> batch.cumsum_along_seq_()
            >>> batch
            tensor([[ 0,  1,  3,  6, 10],
                    [ 5, 11, 18, 26, 35]], batch_dim=0, seq_dim=1)
        """
        self.cumsum_(self._seq_dim)

    def logcumsumexp_along_seq(self) -> BatchedTensorSeq:
        r"""Computes the logarithm of the cumulative summation of the
        exponentiation of elements of the current batch in the sequence
        dimension.

        Returns:
        -------
            ``BatchedTensorSeq``: A batch with the cumulative
                summation of the exponentiation of elements of the
                current batch in the sequence dimension.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> BatchedTensorSeq(torch.arange(10).view(2, 5).float()).logcumsumexp_along_seq()
            tensor([[0.0000, 1.3133, 2.4076, 3.4402, 4.4519],
                    [5.0000, 6.3133, 7.4076, 8.4402, 9.4519]], batch_dim=0, seq_dim=1)
        """
        return self.logcumsumexp(self._seq_dim)

    def logcumsumexp_along_seq_(self) -> None:
        r"""Computes the logarithm of the cumulative summation of the
        exponentiation of elements of the current batch in the sequence
        dimension.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5).float())
            >>> batch.logcumsumexp_along_seq_()
            >>> batch
            tensor([[0.0000, 1.3133, 2.4076, 3.4402, 4.4519],
                    [5.0000, 6.3133, 7.4076, 8.4402, 9.4519]], batch_dim=0, seq_dim=1)
        """
        self.logcumsumexp_(self._seq_dim)

    def permute_along_seq(self, permutation: Sequence[int] | Tensor) -> BatchedTensorSeq:
        r"""Permutes the data along the sequence dimension.

        Args:
        ----
            permutation (sequence or ``torch.Tensor`` of type long
                and shape ``(dimension,)``): Specifies the permutation
                to use on the data. The dimension of the permutation
                input should be compatible with the shape of the data.

        Returns:
        -------
            ``BatchedTensorSeq``: A new batch with permuted data.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5))
            >>> batch.permute_along_seq([2, 1, 3, 0, 4])
            tensor([[2, 1, 3, 0, 4],
                    [7, 6, 8, 5, 9]], batch_dim=0, seq_dim=1)
        """
        return self.permute_along_dim(permutation, dim=self._seq_dim)

    def permute_along_seq_(self, permutation: Sequence[int] | Tensor) -> None:
        r"""Permutes the data along the sequence dimension.

        Args:
        ----
            permutation (sequence or ``torch.Tensor`` of type long
                and shape ``(dimension,)``): Specifies the permutation
                to use on the data. The dimension of the permutation
                input should be compatible with the shape of the data.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5))
            >>> batch.permute_along_seq_([2, 1, 3, 0, 4])
            >>> batch
            tensor([[2, 1, 3, 0, 4],
                    [7, 6, 8, 5, 9]], batch_dim=0, seq_dim=1)
        """
        self.permute_along_dim_(permutation, dim=self._seq_dim)

    def shuffle_along_seq(self, generator: torch.Generator | None = None) -> BatchedTensorSeq:
        r"""Shuffles the data along the sequence dimension.

        Args:
        ----
            generator (``torch.Generator`` or ``None``, optional):
                Specifies an optional random generator.
                Default: ``None``

        Returns:
        -------
            ``BatchedTensorSeq``:  A new batch with shuffled data.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5))
            >>> batch.shuffle_along_seq()  # doctest:+ELLIPSIS
            tensor([[...]], batch_dim=0, seq_dim=1)
        """
        return self.permute_along_seq(torch.randperm(self.seq_len, generator=generator))

    def shuffle_along_seq_(self, generator: torch.Generator | None = None) -> None:
        r"""Shuffles the data along the sequence dimension.

        Args:
        ----
            generator (``torch.Generator`` or ``None``, optional):
                Specifies an optional random generator.
                Default: ``None``

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5))
            >>> batch.shuffle_along_seq_()
            >>> batch  # doctest:+ELLIPSIS
            tensor([[...]], batch_dim=0, seq_dim=1)
        """
        self.permute_along_seq_(torch.randperm(self.seq_len, generator=generator))

    def sort_along_seq(self, *args, **kwargs) -> torch.return_types.sort:
        r"""Sorts the elements of the batch along the sequence dimension
        in monotonic order by value.

        Args:
        ----
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
        -------
            (``BatchedTensorSeq``, ``BatchedTensorSeq``): A tuple with
                two values:
                    - The first batch contains the batch values sorted
                        along the sequence dimension.
                    - The second batch contains the indices that sort
                        the batch along the sequence dimension.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5))
            >>> batch.sort_along_seq(descending=True)
            torch.return_types.sort(
            values=tensor([[4, 3, 2, 1, 0],
                    [9, 8, 7, 6, 5]], batch_dim=0),
            indices=tensor([[4, 3, 2, 1, 0],
                    [4, 3, 2, 1, 0]], batch_dim=0))
        """
        return self.sort(self._seq_dim, *args, **kwargs)

    ################################################
    #     Mathematical | point-wise operations     #
    ################################################

    #############################################
    #     Mathematical | logical operations     #
    #############################################

    ################################
    #     Reduction operations     #
    ################################

    def amax_along_seq(self, keepdim: bool = False) -> Tensor:
        r"""Computes the maximum along the sequence dimension.

        Args:
        ----
            keepdim (bool, optional): Indicates whether the output
                array has the sequence dimension retained or not.
                Default: ``False``

        Returns:
        -------
            ``torch.Tensor``: The maximum along the sequence dimension.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5))
            >>> batch.amax_along_seq()
            tensor([4, 9])
            >>> batch.amax_along_seq(keepdim=True)
            tensor([[4], [9]])
        """
        return self.amax(dim=self._seq_dim, keepdim=keepdim)

    def amin_along_seq(self, keepdim: bool = False) -> Tensor:
        r"""Computes the minimum along the sequence dimension.

        Args:
        ----
            keepdim (bool, optional): Indicates whether the output
                array has the sequence dimension retained or not.
                Default: ``False``

        Returns:
        -------
            ``torch.Tensor``: The minimum along the sequence dimension.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5))
            >>> batch.amin_along_seq()
            tensor([0, 5])
            >>> batch.amin_along_seq(keepdim=True)
            tensor([[0], [5]])
        """
        return self.amin(dim=self._seq_dim, keepdim=keepdim)

    def argmax_along_seq(self, keepdim: bool = False) -> Tensor:
        r"""Computes the indices of the maximum value along the sequence
        dimension.

        Args:
        ----
            keepdim (bool, optional): Indicates whether the output
                array has the sequence dimension retained or not.
                Default: ``False``

        Returns:
        -------
            ``torch.Tensor``: The indices of the maximum value along
                the sequence dimension.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5))
            >>> batch.argmax_along_seq()
            tensor([4, 4])
            >>> batch.argmax_along_seq(keepdim=True)
            tensor([[4], [4]])
        """
        return self.argmax(dim=self._seq_dim, keepdim=keepdim)

    def argmin_along_seq(self, keepdim: bool = False) -> Tensor:
        r"""Computes the indices of the minimum value along the sequence
        dimension.

        Args:
        ----
            keepdim (bool, optional): Indicates whether the output
                array has the sequence dimension retained or not.
                Default: ``False``

        Returns:
        -------
            ``torch.Tensor``: The indices of the minimum value along
                the sequence dimension.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5))
            >>> batch.argmin_along_seq()
            tensor([0, 0])
            >>> batch.argmin_along_seq(keepdim=True)
            tensor([[0], [0]])
        """
        return self.argmin(dim=self._seq_dim, keepdim=keepdim)

    def max_along_seq(self, *args, **kwargs) -> torch.return_types.max:
        r"""Computes the maximum values along the sequence dimension.

        Args:
        ----
            *args: See the documentation of ``torch.Tensor.max``
            **kwargs: See the documentation of ``torch.Tensor.max``

        Returns:
        -------
            ``torch.return_types.max``: A batch with
                the maximum values along the sequence dimension.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5))
            >>> batch.max_along_seq()
            torch.return_types.max(
            values=tensor([4, 9]),
            indices=tensor([4, 4]))
            >>> batch.max_along_seq(keepdim=True)
            torch.return_types.max(
            values=tensor([[4], [9]]),
            indices=tensor([[4], [4]]))
        """
        return self.max(self._seq_dim, *args, **kwargs)

    def mean_along_seq(self, *args, **kwargs) -> Tensor:
        r"""Computes the mean values along the sequence dimension.

        Args:
        ----
            *args: See the documentation of ``torch.Tensor.mean``
            **kwargs: See the documentation of ``torch.Tensor.mean``

        Returns:
        -------
            ``torch.Tensor``: A tensor with the mean values along
                the sequence dimension.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5).float())
            >>> batch.mean_along_seq()
            tensor([2., 7.])
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5).float())
            >>> batch.mean_along_seq(keepdim=True)
            tensor([[2.], [7.]])
        """
        return self.mean(self._seq_dim, *args, **kwargs)

    def median_along_seq(self, *args, **kwargs) -> torch.return_types.median:
        r"""Computes the median values along the sequence dimension.

        Args:
        ----
            *args: See the documentation of ``torch.Tensor.median``
            **kwargs: See the documentation of ``torch.Tensor.median``

        Returns:
        -------
            ``torch.return_types.median``:  The first tensor will
                be populated with the median values and the second
                tensor, which must have dtype long, with their indices
                in the sequence dimension of input.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5))
            >>> batch.median_along_seq()
            torch.return_types.median(
            values=tensor([2, 7]),
            indices=tensor([2, 2]))
        """
        return self.median(self._seq_dim, *args, **kwargs)

    def min_along_seq(self, *args, **kwargs) -> torch.return_types.min:
        r"""Computes the minimum values along the sequence dimension.

        Args:
        ----
            *args: See the documentation of ``torch.Tensor.min``
            **kwargs: See the documentation of ``torch.Tensor.min``

        Returns:
        -------
            ``torch.return_types.min``: A batch with
                the minimum values along the sequence dimension.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5))
            >>> batch.min_along_seq()
            torch.return_types.min(
            values=tensor([0, 5]),
            indices=tensor([0, 0]))
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5))
            >>> batch.min_along_seq(keepdim=True)
            torch.return_types.min(
            values=tensor([[0], [5]]),
            indices=tensor([[0], [0]]))
        """
        return self.min(self._seq_dim, *args, **kwargs)

    def nanmean_along_seq(self, *args, **kwargs) -> Tensor:
        r"""Computes the mean values along the sequence dimension.

        Args:
        ----
            *args: See the documentation of ``torch.Tensor.nanmean``
            **kwargs: See the documentation of ``torch.Tensor.nanmean``

        Returns:
        -------
            ``torch.Tensor``: A batch with
                the mean values along the sequence dimension.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.tensor([[0, 1, 2, 3, 4], [5, 6, 7, 8, float("nan")]]))
            >>> batch.nanmean_along_seq()
            tensor([2.0000, 6.5000])
            >>> batch.nanmean_along_seq(keepdim=True)
            tensor([[2.0000], [6.5000]])
        """
        return self.nanmean(self._seq_dim, *args, **kwargs)

    def nanmedian_along_seq(self, *args, **kwargs) -> torch.return_types.nanmedian:
        r"""Computes the median values along the sequence dimension.

        Args:
        ----
            *args: See the documentation of ``torch.Tensor.nanmedian``
            **kwargs: See the documentation of ``torch.Tensor.nanmedian``

        Returns:
        -------
            ``torch.return_types.nanmedian``:  The first tensor will
                be populated with the median values and the second
                tensor, which must have dtype long, with their indices
                in the sequence dimension of input.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.tensor([[0, 1, 2, 3, 4], [5, 6, 7, 8, float("nan")]]))
            >>> batch.nanmedian_along_seq()
            torch.return_types.nanmedian(
            values=tensor([2., 6.]),
            indices=tensor([2, 1]))
        """
        return self.nanmedian(self._seq_dim, *args, **kwargs)

    def nansum_along_seq(self, *args, **kwargs) -> Tensor:
        r"""Computes the sum values along the sequence dimension.

        Args:
        ----
            *args: See the documentation of ``torch.Tensor.nansum``
            **kwargs: See the documentation of ``torch.Tensor.nansum``

        Returns:
        -------
            ``torch.Tensor``: A tensor with the sum values along the
                sequence dimension.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.tensor([[0, 1, 2, 3, 4], [5, 6, 7, 8, float("nan")]]))
            >>> batch.nansum_along_seq()
            tensor([10., 26.])
        """
        return self.nansum(self._seq_dim, *args, **kwargs)

    def prod_along_seq(self, *args, **kwargs) -> Tensor:
        r"""Computes the product values along the sequence dimension.

        Args:
        ----
            *args: See the documentation of ``torch.Tensor.prod``
            **kwargs: See the documentation of ``torch.Tensor.prod``

        Returns:
        -------
            ``torch.Tensor``: A batch with
                the product values along the sequence dimension.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.tensor([[1, 2, 3, 4, 5], [6, 7, 8, 9, 1]]))
            >>> batch.prod_along_seq()
            tensor([ 120, 3024])
            >>> batch.prod_along_seq(keepdim=True)
            tensor([[ 120], [3024]])
        """
        return self.prod(self._seq_dim, *args, **kwargs)

    def sum_along_seq(self, *args, **kwargs) -> Tensor:
        r"""Computes the sum values along the sequence dimension.

        Args:
        ----
            *args: See the documentation of ``torch.Tensor.sum``
            **kwargs: See the documentation of ``torch.Tensor.sum``

        Returns:
        -------
            ``torch.Tensor``: A tensor with the sum values along the
                sequence dimension.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5))
            >>> batch.sum_along_seq()
            tensor([10, 35])
        """
        return self.sum(self._seq_dim, *args, **kwargs)

    ##########################################################
    #    Indexing, slicing, joining, mutating operations     #
    ##########################################################

    def align_as(self, other: BatchedTensorSeq) -> BatchedTensorSeq:
        r"""Aligns the current batch with the batch ``other``.

        This method makes sure the batch and sequence dimensions
        are aligned.

        Args:
        ----
            other (``BatchedTensorSeq``): Specifies the batch to use to
                align the current batch.

        Returns:
        -------
            ``BatchedTensorSeq``: The aligned batch.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> # batch-sequence -> sequence-batch
            >>> seq_batch = BatchedTensorSeq(torch.ones(2, 3), batch_dim=1, seq_dim=0)
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5))
            >>> batch.align_as(seq_batch)
            tensor([[0, 5],
                    [1, 6],
                    [2, 7],
                    [3, 8],
                    [4, 9]], batch_dim=1, seq_dim=0)
            >>> # sequence-batch -> batch-sequence
            >>> batch_seq = BatchedTensorSeq(torch.ones(2, 3))
            >>> batch = BatchedTensorSeq.from_seq_batch(torch.arange(10).view(5, 2))
            >>> batch.align_as(batch_seq)
            tensor([[0, 2, 4, 6, 8],
                    [1, 3, 5, 7, 9]], batch_dim=0, seq_dim=1)
        """
        if not isinstance(other, self.__class__):
            raise TypeError(
                f"Incorrect type {type(other)}. No implementation available to `align_as` "
                f"{type(self)} with {type(other)}"
            )
        return self.__class__(
            self._data.permute(  # Align only the batch and sequence dims
                *compute_batch_seq_permutation(
                    num_dims=self._data.dim(),
                    old_batch_dim=self.batch_dim,
                    old_seq_dim=self.seq_dim,
                    new_batch_dim=other.batch_dim,
                    new_seq_dim=other.seq_dim,
                )
            ),
            batch_dim=other.batch_dim,
            seq_dim=other.seq_dim,
        )

    def align_to_batch_seq(self) -> BatchedTensorSeq:
        r"""Aligns the current batch to the batch-sequence format.

        Returns:
        -------
            ``BatchedTensorSeq``: The batch in the batch-sequence
                format.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(5, 2), batch_dim=1, seq_dim=0)
            >>> batch.align_to_batch_seq()
            tensor([[0, 2, 4, 6, 8],
                    [1, 3, 5, 7, 9]], batch_dim=0, seq_dim=1)
        """
        return self.__class__(
            align_to_batch_seq(tensor=self._data, **self._get_kwargs()),
            batch_dim=0,
            seq_dim=1,
        )

    def align_to_seq_batch(self) -> BatchedTensorSeq:
        r"""Aligns the current batch to the sequence-batch format.

        Returns:
        -------
            ``BatchedTensorSeq``: The batch in the sequence-batch format.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5), batch_dim=0, seq_dim=1)
            >>> batch.align_to_seq_batch()
            tensor([[0, 5],
                    [1, 6],
                    [2, 7],
                    [3, 8],
                    [4, 9]], batch_dim=1, seq_dim=0)
        """
        return self.__class__(
            align_to_seq_batch(tensor=self._data, batch_dim=self._batch_dim, seq_dim=self._seq_dim),
            batch_dim=1,
            seq_dim=0,
        )

    def cat_along_seq(
        self, tensors: BatchedTensor | Tensor | Iterable[BatchedTensor | Tensor]
    ) -> BatchedTensorSeq:
        r"""Concatenates the data of the batch(es) to the current batch
        along the sequence dimension and creates a new batch.

        Args:
        ----
            tensors (``BatchedTensorSeq`` or ``torch.Tensor`` or
                ``Iterable``): Specifies the batch(es) to concatenate.

        Returns:
        -------
            ``BatchedTensorSeq``: A batch with the concatenated data
                along the sequence dimension.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.tensor([[0, 1, 2], [4, 5, 6]]))
            >>> batch.cat_along_seq(BatchedTensorSeq(torch.tensor([[10, 11], [12, 13]])))
            tensor([[ 0,  1,  2, 10, 11],
                    [ 4,  5,  6, 12, 13]], batch_dim=0, seq_dim=1)
            >>> batch = BatchedTensorSeq(torch.tensor([[0, 4], [1, 5], [2, 6]]), batch_dim=1, seq_dim=0)
            >>> batch.cat_along_seq(
            ...     [
            ...         BatchedTensorSeq(torch.tensor([[10, 12], [11, 13]]), batch_dim=1, seq_dim=0),
            ...         BatchedTensorSeq(torch.tensor([[20, 22], [21, 23]]), batch_dim=1, seq_dim=0),
            ...     ]
            ... )
            tensor([[ 0,  4],
                    [ 1,  5],
                    [ 2,  6],
                    [10, 12],
                    [11, 13],
                    [20, 22],
                    [21, 23]], batch_dim=1, seq_dim=0)
        """
        return self.cat(tensors, dim=self._seq_dim)

    def cat_along_seq_(
        self, tensors: BatchedTensor | Tensor | Iterable[BatchedTensor | Tensor]
    ) -> None:
        r"""Concatenates the data of the batch(es) to the current batch
        along the sequence dimension.

        In-place version of ``cat_along_seq()``.

        Args:
        ----
            tensors (``BatchedTensor`` or ``torch.Tensor`` or
                ``Iterable``): Specifies the batch(es) to concatenate.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.tensor([[0, 1, 2], [4, 5, 6]]))
            >>> batch.cat_along_seq_(BatchedTensorSeq(torch.tensor([[10, 11], [12, 13]])))
            >>> batch
            tensor([[ 0,  1,  2, 10, 11],
                    [ 4,  5,  6, 12, 13]], batch_dim=0, seq_dim=1)
            >>> batch = BatchedTensorSeq(
            ...     torch.tensor([[0, 4], [1, 5], [2, 6]]),
            ...     batch_dim=1,
            ...     seq_dim=0,
            ... )
            >>> batch.cat_along_seq_(
            ...     [
            ...         BatchedTensorSeq(torch.tensor([[10, 12], [11, 13]]), batch_dim=1, seq_dim=0),
            ...         BatchedTensorSeq(torch.tensor([[20, 22], [21, 23]]), batch_dim=1, seq_dim=0),
            ...     ]
            ... )
            >>> batch
            tensor([[ 0,  4],
                    [ 1,  5],
                    [ 2,  6],
                    [10, 12],
                    [11, 13],
                    [20, 22],
                    [21, 23]], batch_dim=1, seq_dim=0)
        """
        self.cat_(tensors, dim=self._seq_dim)

    def chunk_along_seq(self, chunks: int) -> tuple[BatchedTensorSeq, ...]:
        r"""Splits the batch into chunks along the sequence dimension.

        Args:
        ----
            chunks (int): Specifies the number of chunks.

        Returns:
        -------
            tuple: The batch split into chunks along the sequence
                dimension.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5))
            >>> batch.chunk_along_seq(chunks=3)
            (tensor([[0, 1], [5, 6]], batch_dim=0, seq_dim=1),
             tensor([[2, 3], [7, 8]], batch_dim=0, seq_dim=1),
             tensor([[4], [9]], batch_dim=0, seq_dim=1))
        """
        return self.chunk(chunks, self._seq_dim)

    def index_select_along_seq(self, index: Tensor | Sequence[int]) -> BatchedTensorSeq:
        r"""Slices the batch along the sequence dimension at the given
        indices.

        Args:
        ----
            index (``torch.Tensor`` or list or tuple): Specifies the
                indices to select.

        Returns:
        -------
            ``BatchedTensorSeq``: A new batch sliced along the sequence
                dimension at the given indices.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5))
            >>> batch.index_select_along_seq([2, 4])
            tensor([[2, 4],
                    [7, 9]], batch_dim=0, seq_dim=1)
            >>> batch.index_select_along_seq(torch.tensor([4, 3, 2, 1, 0]))
            tensor([[4, 3, 2, 1, 0],
                    [9, 8, 7, 6, 5]], batch_dim=0, seq_dim=1)
        """
        return self.index_select(self._seq_dim, index)

    def repeat_along_seq(self, repeats: int) -> BatchedTensorSeq:
        r"""Repeats the batch along the sequence dimension.

        Args:
        ----
            repeats (int): Specifies the number of times to repeat
                the batch along the sequence dimension.

        Returns:
        -------
            ``BatchedTensorSeq``: A repeated version of the input batch.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5))
            >>> batch.repeat_along_seq(2)
            tensor([[0, 1, 2, 3, 4, 0, 1, 2, 3, 4],
                    [5, 6, 7, 8, 9, 5, 6, 7, 8, 9]], batch_dim=0, seq_dim=1)
        """
        sizes = [1] * self._data.dim()
        sizes[self._seq_dim] = repeats
        return self._create_new_batch(self._data.repeat(*sizes))

    def select_along_seq(self, index: int) -> BatchedTensor:
        r"""Slices the batch along the sequence dimension at the given
        index.

        Args:
        ----
            index (int): Specifies the index to select.

        Returns:
        -------
            ``BatchedTensor``: The batch sliced along the sequence
                dimension at the given index.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5))
            >>> batch.select_along_seq(2)
            tensor([2, 7], batch_dim=0)
        """
        return BatchedTensor(
            data=self._data.select(self._seq_dim, index),
            batch_dim=self._batch_dim if self._seq_dim > self._batch_dim else self._batch_dim - 1,
        )

    def slice_along_seq(
        self, start: int = 0, stop: int | None = None, step: int = 1
    ) -> BatchedTensorSeq:
        r"""Slices the batch in the sequence dimension.

        Args:
        ----
            start (int, optional): Specifies the index where the
                slicing of object starts. Default: ``0``
            stop (int, optional): Specifies the index where the
                slicing of object stops. ``None`` means last.
                Default: ``None``
            step (int, optional): Specifies the increment between
                each index for slicing. Default: ``1``

        Returns:
        -------
            ``BatchedTensorSeq``: A slice of the current batch.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.tensor([[0, 1, 2, 3, 4], [9, 8, 7, 6, 5]]))
            >>> batch.slice_along_seq(start=2)
            tensor([[2, 3, 4],
                    [7, 6, 5]], batch_dim=0, seq_dim=1)
            >>> batch.slice_along_seq(stop=3)
            tensor([[0, 1, 2],
                    [9, 8, 7]], batch_dim=0, seq_dim=1)
            >>> batch.slice_along_seq(step=2)
            tensor([[0, 2, 4],
                    [9, 7, 5]], batch_dim=0, seq_dim=1)
        """
        return self.slice_along_dim(self._seq_dim, start, stop, step)

    def split_along_seq(
        self, split_size_or_sections: int | Sequence[int]
    ) -> tuple[BatchedTensorSeq, ...]:
        return self.split(split_size_or_sections, dim=self._seq_dim)

    def take_along_seq(
        self, indices: BaseBatch | np.ndarray | Tensor | Sequence
    ) -> BatchedTensorSeq:
        r"""Takes values along the sequence dimension.

        Args:
        ----
            indices (``BaseBatch`` or ``numpy.ndarray`` or
                ``torch.Tensor`` or `` Specifies the indices to take
                along the batch dimension.

        Returns:
        -------
            ``BatchedTensorSeq``: The sequence with the selected data.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from redcat import BatchedTensorSeq
            >>> batch = BatchedTensorSeq(torch.arange(10).view(2, 5))
            >>> batch.take_along_seq(BatchedTensorSeq(torch.tensor([[3, 0, 1], [2, 3, 4]])))
            tensor([[3, 0, 1],
                    [7, 8, 9]], batch_dim=0, seq_dim=1)
        """
        return self.take_along_dim(indices, dim=self._seq_dim)

    def unsqueeze(self, dim: int) -> BatchedTensorSeq:
        return self.__class__(
            self._data.unsqueeze(dim=dim),
            batch_dim=self._batch_dim + 1
            if self._batch_dim >= dim and dim >= 0
            else self._batch_dim,
            seq_dim=self._seq_dim + 1 if self._seq_dim >= dim and dim >= 0 else self._seq_dim,
        )

    def _check_valid_dims(self, tensors: Sequence) -> None:
        check_batch_dims(get_batch_dims(tensors))
        check_seq_dims(get_seq_dims(tensors))

    def _get_kwargs(self) -> dict:
        return {"batch_dim": self._batch_dim, "seq_dim": self._seq_dim}


def check_data_and_dims(data: Tensor, batch_dim: int, seq_dim: int) -> None:
    r"""Checks if the tensor ``data``, ``batch_dim`` and ``seq_dim`` are
    correct.

    Args:
    ----
        data (``torch.Tensor``): Specifies the tensor in the batch.
        batch_dim (int): Specifies the batch dimension in the
            ``torch.Tensor`` object.
        seq_dim (int, optional): Specifies the sequence dimension in
            the ``torch.Tensor`` object.

    Raises:
    ------
        RuntimeError: if one of the input is incorrect.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from redcat.tensorseq import check_data_and_dims
        >>> check_data_and_dims(torch.ones(2, 3), batch_dim=0, seq_dim=1)
    """
    if data.dim() < 2:
        raise RuntimeError(f"data needs at least 2 dimensions (received: {data.dim()})")
    if batch_dim < 0 or batch_dim >= data.dim():
        raise RuntimeError(
            f"Incorrect batch_dim ({batch_dim}) but the value should be in [0, {data.dim() - 1}]"
        )
    if seq_dim < 0 or seq_dim >= data.dim():
        raise RuntimeError(
            f"Incorrect seq_dim ({seq_dim}) but the value should be in [0, {data.dim() - 1}]"
        )
    if batch_dim == seq_dim:
        raise RuntimeError(f"batch_dim ({batch_dim}) and seq_dim ({seq_dim}) have to be different")


# def implements(torch_function: Callable) -> Callable:
#     """Register a torch function override for BatchedTensor."""
#
#     def decorator(func: Callable) -> Callable:
#         functools.update_wrapper(func, torch_function)
#         HANDLED_FUNCTIONS[torch_function] = func
#         return func
#
#     return decorator


def from_sequences(
    sequences: Iterable[torch.Tensor], padding_value: bool | int | float = 0
) -> BatchedTensorSeq:
    r"""Converts variable length sequences to a single padded tensor.

    Args:
    ----
        sequences (iterable): Specifies an iterable over the variable
            length sequences. Each sequence is a ``torch.Tensor`` of
            shape ``(sequence_length, *)``. This function assumes
            trailing dimensions and type of all the tensors in
            sequences are same.
        padding_value (bool or int or float, optional): Specifies the
        padding value. Default: ``0``

    Returns:
    -------
        ``BatchedTensorSeq``: A padded tensor. The underlying data is
            a ``torch.Tensor`` of shape
            ``(batch_size, sequence_length, *)``.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from redcat.tensorseq import from_sequences
        >>> from_sequences([torch.ones(3), torch.ones(5), torch.ones(1), torch.ones(0)])
        tensor([[1., 1., 1., 0., 0.],
                [1., 1., 1., 1., 1.],
                [1., 0., 0., 0., 0.],
                [0., 0., 0., 0., 0.]], batch_dim=0, seq_dim=1)
    """
    return BatchedTensorSeq(
        pad_sequence(list(sequences), padding_value=padding_value, batch_first=True)
    )



__all__ = ["BatchExtender", "BatchShuffler", "MiniBatcher"]

from redcat.datapipes.iter.batching import MiniBatcherIterDataPipe as MiniBatcher
from redcat.datapipes.iter.joining import BatchExtenderIterDataPipe as BatchExtender
from redcat.datapipes.iter.shuffling import BatchShufflerIterDataPipe as BatchShuffler



from __future__ import annotations

__all__ = [
    "check_batch_dims",
    "check_data_and_dim",
    "check_seq_dims",
    "get_batch_dims",
    "get_seq_dims",
    "swap2",
]

import copy
from collections.abc import Iterable, Mapping, MutableSequence
from typing import Any, overload

import numpy as np
from numpy import ndarray
from torch import Tensor


def check_batch_dims(dims: set[int]) -> None:
    r"""Gets the batch dimensions from the inputs.

    Args:
    ----
        dims (set): Specifies the batch dims to check.

    Raises:
    ------
        RuntimeError if there are more than one batch dimension.

    Example usage:

    .. code-block:: pycon

        >>> from redcat.utils.common import check_batch_dims
        >>> check_batch_dims({0})
    """
    if len(dims) != 1:
        raise RuntimeError(f"The batch dimensions do not match. Received multiple values: {dims}")


def check_data_and_dim(data: ndarray | Tensor, batch_dim: int) -> None:
    r"""Checks if the array ``data`` and ``batch_dim`` are correct.

    Args:
    ----
        data ( ``numpy.ndarray`` or ``torch.Tensor``): Specifies
            the array in the batch.
        batch_dim (int): Specifies the batch dimension in the
            array object.

    Raises:
    ------
        RuntimeError: if one of the input is incorrect.

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import torch
        >>> from redcat.utils.common import check_data_and_dim
        >>> check_data_and_dim(np.ones((2, 3)), batch_dim=0)
        >>> check_data_and_dim(torch.ones(2, 3), batch_dim=0)
    """
    ndim = data.ndim
    if ndim < 1:
        raise RuntimeError(f"data needs at least 1 dimensions (received: {ndim})")
    if batch_dim < 0 or batch_dim >= ndim:
        raise RuntimeError(
            f"Incorrect batch_dim ({batch_dim}) but the value should be in [0, {ndim - 1}]"
        )


def check_seq_dims(dims: set[int]) -> None:
    r"""Gets the sequence dimensions from the inputs.

    Args:
    ----
        dims (set): Specifies the sequence dims to check.

    Raises:
    ------
        RuntimeError if there are more than one sequence dimension.

    Example usage:

    .. code-block:: pycon

        >>> from redcat.utils.common import get_seq_dims
        >>> get_seq_dims({1})
    """
    if len(dims) != 1:
        raise RuntimeError(
            f"The sequence dimensions do not match. Received multiple values: {dims}"
        )


def get_batch_dims(args: Iterable[Any], kwargs: Mapping[str, Any] | None = None) -> set[int]:
    r"""Gets the batch dimensions from the inputs.

    Args:
    ----
        args: Variable length argument list.
        kwargs: Arbitrary keyword arguments.

    Returns:
    -------
        set: The batch dimensions.

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import torch
        >>> from redcat import BatchedArray, BatchedTensor
        >>> from redcat.utils.common import get_batch_dims
        >>> get_batch_dims(
        ...     args=(BatchedArray(torch.ones(2, 3)), BatchedArray(torch.ones(2, 6))),
        ...     kwargs={"batch": BatchedArray(torch.ones(2, 4))},
        ... )
        {0}
        >>> get_batch_dims(
        ...     args=(BatchedTensor(torch.ones(2, 3)), BatchedTensor(torch.ones(2, 6))),
        ...     kwargs={"batch": BatchedTensor(torch.ones(2, 4))},
        ... )
        {0}
    """
    kwargs = kwargs or {}
    dims = {val._batch_dim for val in args if hasattr(val, "_batch_dim")}
    dims.update({val._batch_dim for val in kwargs.values() if hasattr(val, "_batch_dim")})
    return dims


def get_seq_dims(args: Iterable[Any, ...], kwargs: Mapping[str, Any] | None = None) -> set[int]:
    r"""Gets the sequence dimensions from the inputs.

    Args:
    ----
        args: Variable length argument list.
        kwargs: Arbitrary keyword arguments.

    Returns:
    -------
        set: The sequence dimensions.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from redcat import BatchedTensorSeq
        >>> from redcat.utils.common import get_seq_dims
        >>> get_seq_dims(
        ...     args=(BatchedTensorSeq(torch.ones(2, 3)), BatchedTensorSeq(torch.ones(2, 6))),
        ...     kwargs={"batch": BatchedTensorSeq(torch.ones(2, 4))},
        ... )
        {1}
    """
    kwargs = kwargs or {}
    dims = {val._seq_dim for val in args if hasattr(val, "_seq_dim")}
    dims.update({val._seq_dim for val in kwargs.values() if hasattr(val, "_seq_dim")})
    return dims


@overload
def swap2(sequence: Tensor, index0: int, index1: int) -> Tensor:
    r"""``swap2`` for a ``torch.Tensor``."""


@overload
def swap2(sequence: np.ndarray, index0: int, index1: int) -> np.ndarray:
    r"""``swap2`` for a ``numpy.ndarray``."""


@overload
def swap2(sequence: MutableSequence, index0: int, index1: int) -> MutableSequence:
    r"""``swap2`` for a mutable sequence."""


def swap2(
    sequence: Tensor | np.ndarray | MutableSequence, index0: int, index1: int
) -> Tensor | np.ndarray | MutableSequence:
    r"""Swaps two values in a mutable sequence.

    The swap is performed in-place.

    Args:
    ----
        sequence (``torch.Tensor`` or ``numpy.ndarray`` or
            ``MutableSequence``): Specifies the sequence to update.
        index0 (int): Specifies the index of the first value to swap.
        index1 (int): Specifies the index of the second value to swap.

    Returns:
    -------
        ``torch.Tensor`` or ``numpy.ndarray`` or ``MutableSequence``:
            The updated sequence.

    Example usage:

    .. code-block:: pycon

        >>> from redcat.utils.common import swap2
        >>> seq = [1, 2, 3, 4, 5]
        >>> swap2(seq, 2, 0)
        >>> seq
        [3, 2, 1, 4, 5]
    """
    tmp = copy.deepcopy(sequence[index0])
    sequence[index0] = sequence[index1]
    sequence[index1] = tmp
    return sequence



from __future__ import annotations

__all__ = ["IndexType", "IndicesType", "RNGType"]

import random
from collections.abc import Sequence
from typing import TypeVar, Union
from unittest.mock import Mock

from coola.types import Tensor, ndarray
from coola.utils import is_numpy_available, is_torch_available

if is_numpy_available():
    import numpy as np
else:  # pragma: no cover
    np = Mock()


if is_torch_available():
    import torch
else:  # pragma: no cover
    torch = Mock()

T = TypeVar("T")

IndexType = Union[int, slice, list[int], Tensor, None]

IndicesType = Union[Sequence[int], Tensor, ndarray]

RNGType = Union[random.Random, np.random.Generator, torch.Generator]



from __future__ import annotations

__all__ = ["BatchExtenderIterDataPipe", "create_large_batch"]

import logging
from collections.abc import Iterator, Sequence
from typing import TypeVar

from coola.utils.format import str_indent
from torch.utils.data import IterDataPipe

from redcat.base import BaseBatch

logger = logging.getLogger(__name__)

T = TypeVar("T")


class BatchExtenderIterDataPipe(IterDataPipe[BaseBatch[T]]):
    r"""Implements a DataPipe to combine several ``BaseBatch`` object
    into a single ``BaseBatch`` object.

    Args:
    ----
        datapipe (``IterDataPipe``): Specifies the source
            DataPipe. The DataPipe has to return compatible
            ``BaseBatch`` objects.
        buffer_size (int, optional): Specifies the buffer size i.e.
            the number of batches that are combined into a bigger
            batch. Default: ``10``
        drop_last (bool, optional): If ``True``, the last samples are
            dropped if the buffer is not full, otherwise it is
            returned. Default: ``False``

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from torch.utils.data.datapipes.iter import IterableWrapper
        >>> from redcat import BatchedTensor
        >>> from redcat.datapipes.iter import BatchExtender
        >>> datapipe = BatchExtender(
        ...     IterableWrapper([BatchedTensor(torch.ones(2) * i) for i in range(10)]),
        ...     buffer_size=4,
        ... )
        >>> list(datapipe)
        [tensor([0., 0., 1., 1., 2., 2., 3., 3.], batch_dim=0),
         tensor([4., 4., 5., 5., 6., 6., 7., 7.], batch_dim=0),
         tensor([8., 8., 9., 9.], batch_dim=0)]
        >>> datapipe = BatchExtender(
        ...     IterableWrapper([BatchedTensor(torch.ones(2) * i) for i in range(10)]),
        ...     buffer_size=4,
        ...     drop_last=True,
        ... )
        >>> list(datapipe)
        [tensor([0., 0., 1., 1., 2., 2., 3., 3.], batch_dim=0),
         tensor([4., 4., 5., 5., 6., 6., 7., 7.], batch_dim=0)]
    """

    def __init__(
        self,
        datapipe: IterDataPipe[BaseBatch[T]],
        buffer_size: int = 10,
        drop_last: bool = False,
    ) -> None:
        self._datapipe = datapipe
        if buffer_size < 1:
            raise ValueError(
                f"buffer_size should be greater or equal to 1 (received: {buffer_size})"
            )
        self._buffer_size = int(buffer_size)
        self._drop_last = bool(drop_last)

    def __iter__(self) -> Iterator[BaseBatch[T]]:
        buffer: list[BaseBatch[T]] = []
        for batch in self._datapipe:
            buffer.append(batch)
            if len(buffer) == self._buffer_size:
                yield create_large_batch(buffer)
                buffer = []
        if buffer and not self._drop_last:
            yield create_large_batch(buffer)

    def __len__(self) -> int:
        length = len(self._datapipe)
        if self._drop_last:
            return length // self._buffer_size
        return (length + self._buffer_size - 1) // self._buffer_size

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  buffer_size={self._buffer_size:,},\n"
            f"  drop_last={self._drop_last},\n"
            f"  datapipe={str_indent(self._datapipe)},\n)"
        )


def create_large_batch(batches: Sequence[BaseBatch[T]]) -> BaseBatch[T]:
    r"""Creates a large batch from a sequence of batches.

    Args:
        batches: Specifies the sequence of batches.

    Returns:
        ``BaseBatch``: A batch containing all the examples from the input batches.
    """
    batch = batches[0].clone()  # Create a deepcopy to not change the data in the batch
    batch.extend(batches[1:])
    return batch



from __future__ import annotations

__all__ = ["BatchShufflerIterDataPipe"]

import logging
from collections.abc import Iterator
from typing import TypeVar

from coola.utils.format import str_indent
from torch.utils.data import IterDataPipe

from redcat.base import BaseBatch
from redcat.utils.tensor import get_torch_generator

logger = logging.getLogger(__name__)

T = TypeVar("T")


class BatchShufflerIterDataPipe(IterDataPipe[BaseBatch[T]]):
    r"""Implements a DataPipe to shuffle data in ``BaseBatch`` objects.

    Args:
    ----
        datapipe (``IterDataPipe``): Specifies the source DataPipe.
            The DataPipe has to return ``BaseBatch`` objects.
        random_seed (int, optional): Specifies the random seed used
            to shuffle the data. Default: ``3770589329299158004``

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from torch.utils.data.datapipes.iter import IterableWrapper
        >>> from redcat import BatchedTensor
        >>> from redcat.datapipes.iter import BatchShuffler
        >>> datapipe = BatchShuffler(
        ...     IterableWrapper([BatchedTensor(torch.arange(4).add(i)) for i in range(2)])
        ... )
        >>> list(datapipe)
        [tensor([3, 0, 1, 2], batch_dim=0), tensor([1, 3, 4, 2], batch_dim=0)]
    """

    def __init__(
        self,
        datapipe: IterDataPipe[BaseBatch[T]],
        random_seed: int = 3770589329299158004,
    ) -> None:
        self._datapipe = datapipe
        self._generator = get_torch_generator(random_seed)

    def __iter__(self) -> Iterator[BaseBatch[T]]:
        for batch in self._datapipe:
            yield batch.shuffle_along_batch(self._generator)

    def __len__(self) -> int:
        return len(self._datapipe)

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  random_seed={self.random_seed},\n"
            f"  datapipe={str_indent(self._datapipe)},\n)"
        )

    @property
    def random_seed(self) -> int:
        r"""``int``: The random seed used to initialize the pseudo
        random generator."""
        return self._generator.initial_seed()



from __future__ import annotations

__all__ = [
    "DeviceType",
    "align_to_batch_first",
    "align_to_batch_seq",
    "align_to_seq_batch",
    "compute_batch_seq_permutation",
    "get_available_devices",
    "get_torch_generator",
    "permute_along_dim",
    "to_tensor",
]

from collections.abc import Sequence
from typing import Union

import numpy as np
import torch
from coola.utils.tensor import is_cuda_available, is_mps_available
from torch import Tensor

from redcat.base import BaseBatch
from redcat.utils.common import swap2

DeviceType = Union[torch.device, str, int]


def align_to_batch_first(tensor: Tensor, batch_dim: int) -> Tensor:
    r"""Aligns the input tensor format to ``(batch_size, *)`` where `*`
    means any number of dimensions.

    Args:
    ----
        tensor (``torch.Tensor``): Specifies the tensor to change
            format.
        batch_dim (int): Specifies the batch dimension in the input
            tensor.

    Returns:
    -------
        ``torch.Tensor``: A tensor of shape ``(batch_size, *)`` where
            `*` means any number of dimensions.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from redcat.utils.tensor import align_to_batch_first
        >>> align_to_batch_first(torch.arange(20).view(4, 5), batch_dim=1)
        tensor([[ 0,  5, 10, 15],
                [ 1,  6, 11, 16],
                [ 2,  7, 12, 17],
                [ 3,  8, 13, 18],
                [ 4,  9, 14, 19]])
    """
    if batch_dim == 0:
        return tensor
    return tensor.transpose(0, batch_dim)


def align_to_batch_seq(tensor: Tensor, batch_dim: int, seq_dim: int) -> Tensor:
    r"""Aligns the input tensor format to ``(batch_size, sequence_length,
    *)`` where `*` means any number of dimensions.

    Args:
    ----
        tensor (``torch.Tensor``): Specifies the tensor to change
            format.
        batch_dim (int): Specifies the batch dimension in the input
            tensor.
        seq_dim (int): Specifies the sequence dimension in the input
            tensor.

    Returns:
    -------
        ``torch.Tensor``: A tensor of shape
            ``(batch_size, sequence_length, *)`` where `*` means any
            number of dimensions.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from redcat.utils.tensor import align_to_batch_seq
        >>> align_to_batch_seq(torch.arange(20).view(4, 5), batch_dim=1, seq_dim=0)
        tensor([[ 0,  5, 10, 15],
                [ 1,  6, 11, 16],
                [ 2,  7, 12, 17],
                [ 3,  8, 13, 18],
                [ 4,  9, 14, 19]])
    """
    return tensor.permute(
        *compute_batch_seq_permutation(
            num_dims=tensor.dim(),
            old_batch_dim=batch_dim,
            old_seq_dim=seq_dim,
            new_batch_dim=0,
            new_seq_dim=1,
        )
    )


def align_to_seq_batch(tensor: Tensor, batch_dim: int, seq_dim: int) -> Tensor:
    r"""Aligns the input tensor format to ``(sequence_length, batch_size,
    *)`` where `*` means any number of dimensions.

    Args:
    ----
        tensor (``torch.Tensor``): Specifies the tensor to change
            format.
        batch_dim (int): Specifies the batch dimension in the input
            tensor.
        seq_dim (int): Specifies the sequence dimension in the input
            tensor.

    Returns:
    -------
        ``torch.Tensor``: A tensor of shape
            ``(sequence_length, batch_size, *)`` where `*` means any
            number of dimensions.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from redcat.utils.tensor import align_to_seq_batch
        >>> align_to_seq_batch(torch.arange(20).view(4, 5), batch_dim=0, seq_dim=1)
        tensor([[ 0,  5, 10, 15],
                [ 1,  6, 11, 16],
                [ 2,  7, 12, 17],
                [ 3,  8, 13, 18],
                [ 4,  9, 14, 19]])
    """
    return tensor.permute(
        *compute_batch_seq_permutation(
            num_dims=tensor.dim(),
            old_batch_dim=batch_dim,
            old_seq_dim=seq_dim,
            new_batch_dim=1,
            new_seq_dim=0,
        )
    )


def compute_batch_seq_permutation(
    num_dims: int,
    old_batch_dim: int,
    old_seq_dim: int,
    new_batch_dim: int,
    new_seq_dim: int,
) -> list[int]:
    r"""Computes the permutation to update the batch and sequence
    dimensions.

    Args:
    ----
        num_dims (int): Specifies the number of dimensions.
        old_batch_dim (int): Specifies the old batch dimension.
        old_seq_dim (int): Specifies the old sequence dimension.
        new_batch_dim (int): Specifies the new batch dimension.
        new_seq_dim (int): Specifies the new sequence dimension.

    Returns:
    -------
        list: The permutation to update the batch and sequence
            dimensions.

    Example usage:

    .. code-block:: pycon

        >>> from redcat.utils.tensor import compute_batch_seq_permutation
        >>> compute_batch_seq_permutation(5, 0, 1, 1, 0)
        [1, 0, 2, 3, 4]
        >>> compute_batch_seq_permutation(2, 0, 1, 1, 0)
        [1, 0]
        >>> compute_batch_seq_permutation(5, 0, 1, 2, 0)
        [1, 2, 0, 3, 4]
        >>> compute_batch_seq_permutation(5, 0, 1, 1, 2)
        [2, 0, 1, 3, 4]
    """
    if old_batch_dim == old_seq_dim:
        raise RuntimeError(
            f"Incorrect old_batch_dim ({old_batch_dim}) and old_seq_dim ({old_seq_dim}). "
            "The dimensions should be different"
        )
    if new_batch_dim == new_seq_dim:
        raise RuntimeError(
            f"Incorrect new_batch_dim ({new_batch_dim}) and new_seq_dim ({new_seq_dim}). "
            "The dimensions should be different"
        )
    dims = list(range(num_dims))
    swap2(dims, old_batch_dim, new_batch_dim)  # Swap batch dim
    if old_batch_dim == new_seq_dim and old_seq_dim == new_batch_dim:
        return dims  # Swapping batch dims also swaps sequence dims
    if new_batch_dim == old_seq_dim:
        # Update the old sequence dim because it changes during the batch dim swap
        old_seq_dim = old_batch_dim
    swap2(dims, old_seq_dim, new_seq_dim)  # Swap sequence dim
    return dims


def get_available_devices() -> tuple[str, ...]:
    r"""Gets the available PyTorch devices on the machine.

    Returns
    -------
        tuple: The available devices.

    Example usage:

    .. code-block:: pycon

        >>> from redcat.utils.tensor import get_available_devices
        >>> get_available_devices()  # doctest:+ELLIPSIS
        ('cpu'...)
    """
    devices = ["cpu"]
    if is_cuda_available():
        devices.append("cuda:0")
    if is_mps_available():
        devices.append("mps:0")
    return tuple(devices)


def get_torch_generator(
    random_seed: int = 1, device: torch.device | str | None = "cpu"
) -> torch.Generator:
    r"""Creates a ``torch.Generator`` initialized with a given seed.

    Args:
    ----
        random_seed (int, optional): Specifies a random seed.
            Default: ``1``
        device (``torch.device`` or str or ``None``, optional):
            Specifies the desired device for the generator.
            Default: ``'cpu'``

    Returns:
    -------
        ``torch.Generator``

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from redcat.utils.tensor import get_torch_generator
        >>> generator = get_torch_generator(42)
        >>> torch.rand(2, 4, generator=generator)  # doctest:+ELLIPSIS
        tensor([[...]])
        >>> generator = get_torch_generator(42)
        >>> torch.rand(2, 4, generator=generator)  # doctest:+ELLIPSIS
        tensor([[...]])
    """
    generator = torch.Generator(device)
    generator.manual_seed(random_seed)
    return generator


def permute_along_dim(tensor: Tensor, permutation: Tensor, dim: int = 0) -> Tensor:
    r"""Permutes the values of a tensor along a given dimension.

    Args:
    ----
        tensor (``torch.Tensor``): Specifies the tensor to permute.
        permutation (``torch.Tensor`` of type long and shape
            ``(dimension,)``): Specifies the permutation to use on the
            tensor. The dimension of this tensor should be compatible
            with the shape of the tensor to permute.
        dim (int, optional): Specifies the dimension used to permute the
            tensor. Default: ``0``

    Returns:
    -------
        ``torch.Tensor``: The permuted tensor.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from redcat.utils.tensor import permute_along_dim
        >>> permute_along_dim(tensor=torch.arange(4), permutation=torch.tensor([0, 2, 1, 3]))
        tensor([0, 2, 1, 3])
        >>> permute_along_dim(
        ...     tensor=torch.arange(20).view(4, 5),
        ...     permutation=torch.tensor([0, 2, 1, 3]),
        ... )
        tensor([[ 0,  1,  2,  3,  4],
                [10, 11, 12, 13, 14],
                [ 5,  6,  7,  8,  9],
                [15, 16, 17, 18, 19]])
        >>> permute_along_dim(
        ...     tensor=torch.arange(20).view(4, 5),
        ...     permutation=torch.tensor([0, 4, 2, 1, 3]),
        ...     dim=1,
        ... )
        tensor([[ 0,  4,  2,  1,  3],
                [ 5,  9,  7,  6,  8],
                [10, 14, 12, 11, 13],
                [15, 19, 17, 16, 18]])
        >>> permute_along_dim(
        ...     tensor=torch.arange(20).view(2, 2, 5),
        ...     permutation=torch.tensor([0, 4, 2, 1, 3]),
        ...     dim=2,
        ... )
        tensor([[[ 0,  4,  2,  1,  3],
                 [ 5,  9,  7,  6,  8]],
                [[10, 14, 12, 11, 13],
                 [15, 19, 17, 16, 18]]])
    """
    return tensor.transpose(0, dim)[permutation].transpose(0, dim).contiguous()


def to_tensor(data: BaseBatch | Sequence | Tensor | np.ndarray) -> Tensor:
    r"""Converts the input to a ``torch.Tensor``.

    Args:
    ----
        data (``BaseBatch`` or ``Sequence`` or ``torch.Tensor`` or
            ``numpy.ndarray``): Specifies the data to convert to a
            tensor.

    Returns:
    -------
        ``torch.Tensor``: A tensor.

    Example usage:

    .. code-block:: pycon

        >>> from redcat.utils.tensor import to_tensor
        >>> x = to_tensor([1, 2, 3, 4, 5])
        >>> x
        tensor([1, 2, 3, 4, 5])
    """
    if isinstance(data, BaseBatch):
        data = data.data
    if not torch.is_tensor(data):
        data = torch.as_tensor(data)
    return data



from __future__ import annotations

__all__ = ["to_list"]

from numpy import ndarray
from torch import Tensor

from redcat import BaseBatch


def to_list(data: list | tuple | Tensor | ndarray | BaseBatch) -> list:
    r"""Converts an input data to a list.

    Args:
    ----
        data (list or tuple or ``torch.Tensor`` or ``numpy.ndarray``
            or ``BaseBatch``): Specifies the data to convert.

    Returns:
    -------
        list: The data.

     Example usage:

    .. code-block:: pycon

        >>> from redcat import BatchList
        >>> from redcat.utils.collection import to_list
        >>> to_list(BatchList([1, 2, 3]))
        [1, 2, 3]
    """
    if isinstance(data, list):
        return data
    if isinstance(data, BaseBatch):
        data = data.data
    if isinstance(data, (Tensor, ndarray)):
        return data.tolist()
    return list(data)



from __future__ import annotations

__all__ = ["permute_along_dim", "to_array"]

from collections.abc import Sequence

import numpy as np
import torch
from numpy import ndarray

from redcat.base import BaseBatch


def permute_along_dim(array: ndarray, permutation: ndarray, dim: int = 0) -> ndarray:
    r"""Permutes the values of a array along a given dimension.

    Args:
    ----
        array (``numpy.ndarray``): Specifies the array to permute.
        permutation (``numpy.ndarray`` of type long and shape
            ``(dimension,)``): Specifies the permutation to use on the
            array. The dimension of this array should be compatible
            with the shape of the array to permute.
        dim (int, optional): Specifies the dimension used to permute the
            array. Default: ``0``

    Returns:
    -------
        ``numpy.ndarray``: The permuted array.

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> from redcat.utils.array import permute_along_dim
        >>> permute_along_dim(np.arange(4), permutation=np.array([0, 2, 1, 3]))
        array([0, 2, 1, 3])
        >>> permute_along_dim(
        ...     np.arange(20).reshape(4, 5),
        ...     permutation=np.array([0, 2, 1, 3]),
        ... )
        array([[ 0,  1,  2,  3,  4],
               [10, 11, 12, 13, 14],
               [ 5,  6,  7,  8,  9],
               [15, 16, 17, 18, 19]])
        >>> permute_along_dim(
        ...     np.arange(20).reshape(4, 5),
        ...     permutation=np.array([0, 4, 2, 1, 3]),
        ...     dim=1,
        ... )
        array([[ 0,  4,  2,  1,  3],
               [ 5,  9,  7,  6,  8],
               [10, 14, 12, 11, 13],
               [15, 19, 17, 16, 18]])
        >>> permute_along_dim(
        ...     np.arange(20).reshape(2, 2, 5),
        ...     permutation=np.array([0, 4, 2, 1, 3]),
        ...     dim=2,
        ... )
        array([[[ 0,  4,  2,  1,  3],
                [ 5,  9,  7,  6,  8]],
               [[10, 14, 12, 11, 13],
                [15, 19, 17, 16, 18]]])
    """
    return np.swapaxes(np.swapaxes(array, 0, dim)[permutation], 0, dim)


def to_array(data: Sequence | torch.Tensor | ndarray) -> ndarray:
    r"""Converts the input to a ``numpy.ndarray``.

    Args:
    ----
        data (``BaseBatch`` or ``Sequence`` or ``torch.Tensor`` or
            ``numpy.ndarray``): Specifies the data to convert to an
            array.

    Returns:
    -------
        ``numpy.ndarray``: An array.

    Example usage:

    .. code-block:: pycon

        >>> from redcat.utils.array import to_array
        >>> x = to_array([1, 2, 3, 4, 5])
        >>> x
        array([1, 2, 3, 4, 5])
    """
    if isinstance(data, BaseBatch):
        data = data.data
    if not isinstance(data, ndarray):
        data = np.asarray(data)
    return data
